# -*- coding: utf-8 -*-
"""
    drift - API key rules
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Implement api key rules for passing, redirecting or rejecting requests based
    on product and version of the client.
"""
from __future__ import absolute_import
import logging
import re

from six.moves import http_client
from six.moves.urllib.parse import urlparse

from flask import request, g, jsonify, make_response, current_app

from driftconfig.relib import ConstraintError


log = logging.getLogger(__name__)


def drift_init_extension(app, **kwargs):
    @app.before_request
    def check_api_key_rules():
        conf = current_app.extensions['driftconfig'].get_config()
        rule = get_api_key_rule(request.headers, request.url, conf)
        if not rule:
            return

        # Update response header in the "after request" callback function further below.
        if 'response_header' in rule:
            g.update_response_header_api_rule = rule['response_header']

        if rule['status_code'] is not None:
            log.info("Blocking request because of api key rule: %s", rule)
            response = make_response(jsonify(rule['response_body']), rule['status_code'])
            return response

    @app.after_request
    def after_request_apply_api_rules(response):
        if hasattr(g, 'update_response_header_api_rule'):
            for k, v in g.update_response_header_api_rule.items():
                response.headers[k] = v

        return response


def get_api_key_rule(request_headers, request_url, conf):
    # Looks up and matches an api key rule to a given key and client version.
    # Returns None if no rule or action is in effect, else a dict with the following
    # optional entries that should be used in response to the http request:
    # 'status_code', 'response_body' and 'response_header'.
    # If 'status_code' is None, the request should be processed further.

    # Apply pass-through rules for legacy keys
    nginx_conf = conf.table_store.get_table('nginx').get(
        {'tier_name': conf.tier['tier_name']})
    if nginx_conf:
        for rule in nginx_conf.get('api_key_passthrough', []):
            key = request_headers.get(rule['key_name'])
            if key and re.match(rule['key_value'], key):
                log.info("Passing on request that contains legacy api key '%s'.", key)
                return

    # We don't require the key to exist in the request header as it is usually
    # enforced by the api router. Some endpoints are actually "keyless". But if the
    # key does exist in the request header, we apply the rules accordingly.
    key = request_headers.get('Drift-Api-Key')
    if not key:
        return

    if ':' in key:
        key, version = key.rsplit(':', 1)
    else:
        version = None

    def retval(status_code=None, message=None, description=None, rule=None):
        status_code = status_code or http_client.FORBIDDEN
        message = message or 'Forbidden'
        description = description or message
        response_body = {
            "error": {
                "code": "user_error",
                "description": description
            },
            "message": message,
            "status_code": status_code
        }

        return {
            'status_code': status_code,
            'response_header': {'Content-Type': 'application/json'},
            'response_body': response_body,
            'rule': rule,
            'api_key': key,
            'api_key_version': version,
        }

    # Fist look up the API key in our config.
    try:
        api_key = conf.table_store.get_table('api-keys').get({'api_key_name': key})
    except ConstraintError:
        return retval(description="API Key format '{}' not recognized.".format(key))

    if not api_key:
        return retval(description="API Key '{}' not found.".format(key))

    # See if the API key is active or not.
    if not api_key['in_use']:
        return retval(description="API Key '{}' is disabled.".format(key))

    # Product keys must point to a product.
    if api_key['key_type'] == 'product' and 'product_name' not in api_key:
        return retval(
            description="API Key type is 'product', but has no reference to any product."
        )

    # If key is not associated with any product, there are no product rules to apply
    # so we are done here.
    if 'product_name' not in api_key:
        return

    # Match the API key to the product/tenant.
    product, tenant = conf.product, conf.tenant
    if not product or not tenant:
        return retval(description="No product or tenant in context.")

    if api_key['product_name'] != product['product_name']:
        return retval(
            description="API Key '{}' is for product '{}'"
            " but current tenant '{}' is on product '{}'.".format(
                key, api_key['product_name'], tenant['tenant_name'], product['product_name']))

    # Now we apply the actual rules for this product.
    rules = conf.table_store.get_table('api-key-rules').find(
        {'product_name': product['product_name'], 'is_active': True})
    rules = sorted(rules, key=lambda rule: rule['assignment_order'])

    for rule in rules:
        patterns = rule['version_patterns']
        is_match = False

        # If no pattern is specified, it effectively means "match always".
        if not patterns:
            is_match = True

        # If the caller provided its version, we use it to match certain rules.
        if version:
            for pattern in patterns:
                if version.startswith(pattern):
                    is_match = True
                    break

        # If the rule did not match the client version, continue to next one.
        if not is_match:
            continue

        ret = retval(rule=rule)
        ret['response_header'].update(rule.get('response_header', {}))

        if rule['rule_type'] == 'pass':
            ret['status_code'] = None  # Signal a pass on this request.
            return ret

        if rule['rule_type'] == 'reject':
            if 'reject' in rule:
                ret['response_body'] = rule['reject']['response_body']
                ret['status_code'] = rule['reject']['status_code']
            else:
                ret['response_body'] = {'message': 'Forbidden'}
                ret['status_code'] = http_client.FORBIDDEN
            return ret

        elif rule['rule_type'] == 'redirect':
            urlparts = urlparse(request_url)
            if '.' not in urlparts.hostname:
                return retval(
                    status_code=400,
                    message='Bad Request',
                    description="Can't redirect to new tenant when hostname is dotless.",
                    rule=rule,
                )

            current_tenant, domain = urlparts.hostname.split('.', 1)

            # See if the host already matches the redirection.
            if current_tenant == rule['redirect']['tenant_name']:
                continue

            redirect = rule['redirect']
            new_hostname = redirect['tenant_name'] + '.' + domain
            url = request_url.replace(urlparts.hostname, new_hostname)
            ret['status_code'] = 307  # Temporary redirect
            ret['response_body'] = {'message': "Redirecting to tier '{}'.".format(redirect['tenant_name'])}
            ret['response_header']['Location'] = url

            return ret



# -*- coding: utf-8 -*-
"""
    drift - tracking setup code
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Setup tracking metadata for the http requests and responses
"""
from __future__ import absolute_import

import uuid
import logging
from flask import request, current_app

CORRELATION_ID = "Correlation-ID"

log = logging.getLogger(__name__)


def drift_init_extension(app, **kwargs):
    @app.before_request
    def add_correlation_id():
        correlation_id = request.headers.get(CORRELATION_ID)
        log.debug("%s %s", request.method, request.url)
        if not correlation_id:
            correlation_id = str(uuid.uuid4())
            if request.method != "GET":
                """
                TODO: remove sensitive information such as username/password
                """
                log.debug({
                    "message": "Tracking request",
                    "correlation_id": correlation_id,
                    "method": request.method,
                    "uri": request.url,
                    "data": request.data,
                })
        request.correlation_id = correlation_id

    @app.after_request
    def add_response_headers(response):
        if CORRELATION_ID not in response.headers:
            response.headers[CORRELATION_ID] = getattr(request, "correlation_id", None)
        response.headers["Host-Address"] = current_app.config.get("HOST_ADDRESS", "Unknown")
        return response



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import os
import datetime
import logging

from six.moves import cPickle as pickle, http_client
import redis

from flask import g, abort
from flask import _app_ctx_stack as stack
from werkzeug._compat import integer_types
from werkzeug.local import LocalProxy

from driftconfig.util import get_parameters
from drift.core.extensions.driftconfig import check_tenant


log = logging.getLogger(__name__)


REDIS_DB = 0  # We always use the main redis db for now


# defaults when making a new tier
TIER_DEFAULTS = {
    "host": "<PLEASE FILL IN>",
    "port": 6379,
    "socket_timeout": 5,
    "socket_connect_timeout": 5
}

HAS_LOCAL_SERVER_MODE = True  # Supports DRIFT_USE_LOCAL_SERVERS flag.


def _get_redis_connection_info():
    """
    Return tenant specific Redis connection info, if available, else use one that's
    specified for the tier.
    """
    ci = None
    if g.conf.tenant:
        ci = g.conf.tenant.get('redis')
    return ci


def provision_resource(ts, tenant_config, attributes):
    """
    Create, recreate or delete resources for a tenant.
    'tenant_config' is a row from 'tenants' table for the particular tenant, tier and deployable.
    LEGACY SUPPORT: 'attributes' points to the current resource attributes within 'tenant_config'.
    """
    report = []
    attributes = attributes.copy()  # Make a copy so we won't modify the actual drift config db.
    if os.environ.get('DRIFT_USE_LOCAL_SERVERS', False):
        attributes['host'] = os.environ.get('DRIFT_REDIS_HOST', 'localhost')

    # Reset Redis cache when initializing or uninitializing.
    if tenant_config['state'] in ['initializing', 'uninitializing']:
        red = RedisCache(tenant_config['tenant_name'], tenant_config['deployable_name'], attributes)
        log.info(
            "Deleting redis cache  on '%s' as the tenant is %s.",
            red.make_key("*"), tenant_config['state']
        )
        red.delete_all()
        report.append("Redis cache was flushed.")
    else:
        report.append("No action needed.")

    return report


class RedisExtension(object):
    """Redis Flask extension."""
    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if not hasattr(app, 'extensions'):
            app.extensions = {}

        app.extensions['redis'] = self
        app.before_request(self.before_request)

    def before_request(self):
        g.redis = LocalProxy(self.get_session)

    def get_session(self):
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'redis_session'):
                ctx.redis_session = get_redis_session()
            return ctx.redis_session

    def get_session_if_available(self):
        """Return redis session if it's already initialized."""
        if stack.top:
            return getattr(stack.top, 'redis_session', None)


@check_tenant
def get_redis_session():
    if g.conf.tenant and g.conf.tenant.get("redis"):
        redis_config = g.conf.tenant.get("redis")
        return RedisCache(
            g.conf.tenant_name['tenant_name'],
            g.conf.deployable['deployable_name'],
            redis_config
        )
    else:
        abort(http_client.BAD_REQUEST, "No Redis resource configured.")


def drift_init_extension(app, **kwargs):
    RedisExtension(app)


class RedisCache(object):
    """
    A wrapper around the redis cache cluster which adds tenancy
    """
    conn = None
    tenant = None
    disabled = False

    def __init__(self, tenant, service_name, redis_config):
        self.disabled = redis_config.get("disabled", False)
        if self.disabled:
            log.warning("Redis is disabled!")
            return

        self.tenant = tenant
        self.service_name = service_name
        self.host = redis_config["host"]
        self.port = redis_config["port"]

        # Override Redis hostname if needed
        if os.environ.get('DRIFT_USE_LOCAL_SERVERS', False):
            self.host = os.environ.get('DRIFT_REDIS_HOST', 'localhost')

        self.conn = redis.StrictRedis(
            host=self.host,
            port=self.port,
            socket_timeout=redis_config.get("socket_timeout", 5),
            socket_connect_timeout=redis_config.get("socket_connect_timeout", 5),
            db=redis_config.get("db_number", REDIS_DB),
            retry_on_timeout=redis_config.get("retry_on_timeout", True),
        )

        self.key_prefix = "{}.{}:".format(self.tenant, self.service_name)

        log.debug("RedisCache initialized. self.conn = %s", self.conn)

    def make_key(self, key):
        """
        Create a redis key with tenant and service embedded into the key
        """
        return self.key_prefix + key

    def dump_object(self, value):
        """Dumps an object into a string for redis.  By default it serializes
        integers as regular string and pickle dumps everything else.
        """
        t = type(value)
        if t in integer_types:
            return str(value).encode('ascii')
        return b'!' + pickle.dumps(value)

    def load_object(self, value):
        """The reversal of :meth:`dump_object`.  This might be called with
        None.
        """
        if value is None:
            return None
        if value.startswith(b'!'):
            try:
                return pickle.loads(value[1:])
            except pickle.PickleError:
                return None
        try:
            return int(value)
        except ValueError:
            return value

    def set(self, key, value, expire=-1):
        compound_key = self.make_key(key)
        dump = self.dump_object(value)
        if expire == -1:
            result = self.conn.set(name=compound_key, value=dump)
        else:
            result = self.conn.setex(name=compound_key, value=dump, time=expire)

        return result

    def get(self, key):
        compound_key = self.make_key(key)
        try:
            ret = self.conn.get(compound_key)
        except redis.RedisError:
            log.exception("Can't fetch key '%s'", compound_key)
            return None

        ret = self.load_object(ret)
        return ret

    def delete(self, key):
        """
        Delete the item with the specified key
        """
        if self.disabled:
            log.info("Redis disabled. Not deleting key '%s'", key)
            return None
        compound_key = self.make_key(key)
        self.conn.delete(compound_key)

    def incr(self, key, amount=1, expire=None):
        """
        Increments the value of 'key' by 'amount'. If no key exists,
        the value will be initialized as 'amount'
        """
        if self.disabled:
            log.info("Redis disabled. Not incrementing key '%s'", key)
            return None
        compound_key = self.make_key(key)
        self.conn.incr(compound_key, amount)
        if expire:
            self.conn.expire(compound_key, expire)

    def lock(self, lock_name):
        return self.conn.lock(self.make_key(lock_name))

    def delete_all(self):
        """remove all the keys for this tenant from redis"""
        if self.disabled:
            log.info("Redis disabled. Not deleting all keys")
            return
        self.set('_dummy', 'ok')
        compound_key = self.make_key("*")
        for key in self.conn.scan_iter(compound_key, count=150000):  # Arbitrary count but otherwise it locks up.
            self.conn.delete(key)


# NOTE THIS IS DEPRECATED AND NEEDS TO BE UPGRADED TO NU STYLE PROVISIONING LOGIC
def provision(config, args, recreate=None):
    params = get_parameters(config, args, TIER_DEFAULTS.keys(), "redis")
    if os.environ.get('DRIFT_USE_LOCAL_SERVERS', False):
        params['host'] = os.environ.get('DRIFT_REDIS_HOST', 'localhost')
    config.tenant["redis"] = params

    if recreate == 'recreate':
        red = RedisCache(config.tenant_name['tenant_name'], config.deployable['deployable_name'], params)
        red.delete_all()


def healthcheck():
    if "redis" not in g.conf.tenant:
        raise RuntimeError("Tenant config does not have 'redis'")
    for k in TIER_DEFAULTS.keys():
        if not g.conf.tenant["redis"].get(k):
            raise RuntimeError("'redis' config missing key '%s'" % k)

    dt = datetime.datetime.utcnow().isoformat()
    g.redis.set("healthcheck", dt, expire=120)
    if g.redis.get("healthcheck") != dt:
        raise RuntimeError("Unexpected redis value")



# -*- coding: utf-8 -*-

import os
import socket
import collections
import platform
import datetime
import json
import logging

from flask import request, current_app, g
from flask import url_for
from flask.views import MethodView
from flask_smorest import Blueprint
import marshmallow as ma
from drift.utils import get_tier_name
import werkzeug.routing
import drift, driftconfig

from drift.core.extensions.jwt import current_user


log = logging.getLogger(__name__)


bp = Blueprint('root', 'Service Status', url_prefix='/', description='Status of the service')


class ServiceStatusSchema(ma.Schema):
    result = ma.fields.Str(description="Is the service healthy")
    host_info = ma.fields.Dict(description="Information about the host machine")
    build_info = ma.fields.Dict(description="Information about the build")
    service_name = ma.fields.Str(description="Name of this service deployable")
    endpoints = ma.fields.Dict(description="List of all exposed endpoints. Contains contextual information if Auth header is provided")
    current_user = ma.fields.Dict(description="Decoded JWT of the current user")
    tier_name = ma.fields.Str(description="Name of this tier")
    tenant_name = ma.fields.Str(description="Name of the current tenant")
    server_time = ma.fields.DateTime(description="Current wallclock time of the server machine")
    tenants = ma.fields.Dict(description="Information about tenants")
    platform = ma.fields.Dict(description="Information about the platform")
    config_dump = ma.fields.Dict(description="Dump of the entire Flask config")
    default_tenant = ma.fields.Str(description="Default tenant name")
    request_headers = ma.fields.Dict(description="Request headers (debug only)")
    request_object = ma.fields.Dict(description="Request object info (debug only)")
    wsgi_env = ma.fields.Dict(description="WSGI Environment")
    version = ma.fields.Str(description="Service version")


def drift_init_extension(app, api, **kwargs):
    api.register_blueprint(bp)


@bp.route('', endpoint='root')
class InfoPageAPI(MethodView):

    no_jwt_check = ["GET"]

    @bp.response(ServiceStatusSchema)
    def get(self):
        """
        Root endpoint

        Basic information about the service and exposed endpoints.
        This is the place to start for interacting with the service.
        """
        tier_name = get_tier_name()
        deployable_name = current_app.config['name']

        # See if caller is privy to some extra info
        show_extra_info = (current_user and ('service' in current_user['roles'])) or current_app.debug

        host_info = collections.OrderedDict()
        host_info["host-name"] = None
        host_info["image"] = current_app.config.get("DOCKER_IMAGE")
        if current_app.config.get("HOST_ADDRESS"):
            host_info["ip-address"] = current_app.config.get("HOST_ADDRESS")
        else:
            host_info["host-name"] = socket.gethostname()
            try:
                host_info["ip-address"] = socket.gethostbyname(
                    socket.gethostname()
                )
            except Exception:
                """
                TODO: this is just a work around
                there might be a better way to get the address
                """
                host_info["ip-address"] = "Unknown"

        # Platform info
        keys = [
            'architecture',
            'machine',
            'node',
            'platform',
            'processor',
            'python_branch',
            'python_build',
            'python_compiler',
            'python_implementation',
            'python_revision',
            'python_version',
            'release',
            'system',
            'version',
        ]
        try:
            platform_info = {key: getattr(platform, key)() for key in keys}

            platform_info["drift_version"] = drift.__version__
            platform_info["driftconfig_version"] = driftconfig.__version__
        except Exception as e:
            platform_info = str(e)
        endpoints = collections.OrderedDict()
        endpoints["root"] = url_for("root.root", _external=True)
        if endpoints["root"].endswith("/"):
            endpoints["root"] = endpoints["root"][:-1]
        for func in current_app.endpoint_registry_funcs + current_app.endpoint_registry_funcs2:
            try:
                endpoints.update(func(current_user))
            except Exception:
                log.exception("Failed to get endpoint registry from %s", func)

        # Publish Swagger and ReDoc if available
        try:
            # <Rule '/doc/openapi.json' (GET, HEAD, OPTIONS) -> api-docs.openapi_json>,
            # <Rule '/doc/swagger' (GET, HEAD, OPTIONS) -> api-docs.openapi_swagger_ui>,
            # <Rule '/doc/redoc' (GET, HEAD, OPTIONS) -> api-docs.openapi_redoc>,
            endpoints["openapi"] = url_for("api-docs.openapi_json", _external=True)
            endpoints["swagger"] = url_for("api-docs.openapi_swagger_ui", _external=True)
            endpoints["redoc"] = url_for("api-docs.openapi_redoc", _external=True)
        except werkzeug.routing.BuildError:
            pass  # Either it's there or it isn't.

        # Only list out tenants which have a db, and only if caller has service role.
        if show_extra_info:
            ts = g.conf.table_store
            tenants_table = ts.get_table('tenants')
            tenants = []
            for tenant in tenants_table.find({'tier_name': tier_name, 'deployable_name': deployable_name}):
                tenants.append(tenant['tenant_name'])

        else:
            tenants = None

        ret = {
            "service_name": current_app.config['name'],
            "version": current_app.config.get('VERSION', "Unknown"),
            "host_info": host_info,
            "build_info": current_app.config.get("BUILD_INFO"),
            "endpoints": endpoints,
            "current_user": dict(current_user) if current_user else None,
            "tier_name": tier_name,
            "tenant_name": g.conf.tenant_name['tenant_name'] if g.conf.tenant_name else '(none)',
            "server_time": datetime.datetime.utcnow(),
            "tenants": tenants,
            "platform": platform_info,
        }

        path = os.path.join(current_app.instance_path, "..", "deployment-manifest.json")
        if not os.path.exists(path):
            if current_app.debug or current_app.testing:
                # Running in debug or testing mode usually means running on local dev machine, which
                # usually means there is no deployment manifest, and no-one should care.
                pass
            else:
                log.debug("No deployment manifest found at %s", path)
        else:
            try:
                ret["deployment"] = json.load(open(path))
            except Exception:
                log.exception("Failed to read deployment manifest from %s", path)

        if show_extra_info:
            # TODO: Only do for authenticated sessions.. preferably..
            ret["request_headers"] = dict(request.headers)
            ret['request_object'] = {
                'remote_addr': request.remote_addr,
                'path': request.path,
                'full_path': request.full_path,
                'script_root': request.script_root,
                'url': request.url,
                'base_url': request.base_url,
                'url_root': request.url_root,
                'authorization': request.authorization,
                'endpoint': request.endpoint,
                'host': request.host,
                'remote_user': request.remote_user,
            }
            ret['wsgi_env'] = {k: str(v) for k, v in request.environ.items()}

            # Pretty print the config
            d = {k: str(v) for k, v in current_app.config.items()}
            d = collections.OrderedDict(sorted(d.items()))
            d['private_key'] = '...'  # Just to be safe(r)
            ret['config_dump'] = json.dumps(d, indent=4)
            ret['default_tenant'] = os.environ.get('DRIFT_DEFAULT_TENANT')

        return ret



# -*- coding: utf-8 -*-
import sys
import unittest
import os.path
import importlib


"""
Test the import of various top level modules to make sure they compile
"""


class ImportTestCase(unittest.TestCase):
    def test_devapp(self):
        importlib.import_module("drift.contrib.flask.devapp")

    def test_fixers(self):
        importlib.import_module("drift.fixers")

    def test_flaskfactory(self):
        importlib.import_module("drift.flaskfactory")

    def test_orm(self):
        importlib.import_module("drift.orm")

    def test_systesthelper(self):
        importlib.import_module("drift.systesthelper")

    def test_urlregistry(self):
        importlib.import_module("drift.core.extensions.urlregistry")

    def test_utils(self):
        importlib.import_module("drift.utils")

    def test_uwsgiboot(self):
        importlib.import_module("drift.contrib.flask.plainapp")

    def test_version(self):
        importlib.import_module("drift.version")

    def test_webservers(self):
        importlib.import_module("drift.webservers")

    def test_drift_commands(self):
        import drift.management
        valid_commands = drift.management.get_commands()
        for cmd in valid_commands:
            importlib.import_module("drift.management.commands." + cmd)


class ScriptImportTestCase(unittest.TestCase):
    script_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "scripts"))

    def setUp(self):
        sys.path.append(self.script_dir)

    def tearDown(self):
        sys.path.pop()

    def test_parse_uwsgi_profiler(self):
        importlib.import_module("parse_uwsgi_profiler")

    def test_travis_build_dependent_projects(self):
        importlib.import_module("travis_build_dependent_projects")



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging

from flask import g


log = logging.getLogger(__name__)


def before_request():
    g.client_debug_messages = []


def after_request(response):
    if hasattr(g, "client_debug_messages") and len(g.client_debug_messages) > 0:
        response.headers["Drift-Debug-Message"] = "\\n".join(g.client_debug_messages)
    return response


def drift_init_extension(app, **kwargs):
    app.before_request(before_request)
    app.after_request(after_request)

    # Install DebugToolbar if applicable
    if app.debug:
        log.info("Flask server is running in DEBUG mode.")
        try:
            from flask_debugtoolbar import DebugToolbarExtension
            DebugToolbarExtension(app)
        except ImportError:
            log.info("Flask DebugToolbar not available: Do 'pip install "
                     "flask-debugtoolbar' to enable.")



# -*- coding: utf-8 -*-
"""
    Adding HTTP Method Overrides
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Some HTTP proxies do not support arbitrary HTTP methods or newer HTTP methods
    (such as PATCH). In that case it’s possible to “proxy” HTTP methods through another
    HTTP method in total violation of the protocol.

    The way this works is by letting the client do an HTTP POST request and set the
    X-HTTP-Method-Override header and set the value to the intended HTTP method (such
    as PATCH).

    http://flask.pocoo.org/docs/0.12/patterns/methodoverrides/
"""
from __future__ import absolute_import


class HTTPMethodOverrideMiddleware(object):
    allowed_methods = frozenset([
        'GET',
        'HEAD',
        'POST',
        'DELETE',
        'PUT',
        'PATCH',
        'OPTIONS'
    ])

    bodyless_methods = frozenset(['GET', 'HEAD', 'OPTIONS', 'DELETE'])

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        method = environ.get('HTTP_X_HTTP_METHOD_OVERRIDE', '').upper()
        if method in self.allowed_methods:
            method = method.encode('ascii', 'replace')
            environ['REQUEST_METHOD'] = method
        if method in self.bodyless_methods:
            environ['CONTENT_LENGTH'] = '0'
        return self.app(environ, start_response)


def drift_init_extension(app, **kwargs):
    app.wsgi_app = HTTPMethodOverrideMiddleware(app.wsgi_app)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import logging
from datetime import datetime, timedelta
import json
from functools import wraps
import re
import string

import jwt
from six.moves.http_client import UNAUTHORIZED

from flask import current_app, request, _request_ctx_stack, g, url_for, redirect, make_response
from flask.views import MethodView
from flask_smorest import Blueprint, abort

import marshmallow as ma

from werkzeug.local import LocalProxy
from werkzeug.security import gen_salt

from drift.utils import get_tier_name
from drift.core.extensions.urlregistry import Endpoints
from drift.core.extensions.tenancy import current_tenant_name, split_host

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_public_key


JWT_VERIFY_CLAIMS = ['signature', 'exp', 'iat']
JWT_REQUIRED_CLAIMS = ['exp', 'iat', 'jti']
JWT_ALGORITHM = 'RS256'
JWT_EXPIRATION_DELTA = 60 * 60 * 24
JWT_LEEWAY = 10

# Tis a hack:
JWT_EXPIRATION_DELTA_FOR_SERVICES = 60 * 60 * 24 * 365

# Implicitly trust following issuers:
TRUSTED_ISSUERS = set(['drift-base'])

# list of regular expression objects matching endpoint definitions
WHITELIST_ENDPOINTS = [
    r"^api-docs\.",  # the marshmalloc documentation endpoint
    r"^static$",  # the marshmalloc documentation endpoint
    ]

SESSION_COOKIE_NAME = 'drift-session'

# List of open endpoints, i.e. not requiring a valid JWT.
# these are the view functions themselves
_open_endpoints = set()

log = logging.getLogger(__name__)
bp = Blueprint('auth', 'Authentication', url_prefix='/auth', description='Authentication endpoints')
bpjwks = Blueprint('jwks', 'JSON Web Key Set', url_prefix='/.well-known')
endpoints = Endpoints()


def drift_init_extension(app, api, **kwargs):
    api.register_blueprint(bp)
    api.register_blueprint(bpjwks)
    endpoints.init_app(app)

    # Flask Secret must be set for cookies and other secret things
    # HACK WARNING: It is weirdly difficult to get a drift config at this point.
    from drift.flaskfactory import load_flask_config
    from driftconfig.util import get_default_drift_config
    ts = get_default_drift_config()
    public_keys = ts.get_table('public-keys')
    row = public_keys.get({
        'tier_name': get_tier_name(),
        'deployable_name': load_flask_config()['name'],
    })

    # If there is no 'secret_key' specified then session cookies are not supported.
    if row and 'secret_key' in row:
        app.config['SECRET_KEY'] = row['secret_key']
        app.config['SESSION_COOKIE_NAME'] = SESSION_COOKIE_NAME

    if not hasattr(app, "jwt_auth_providers"):
        app.jwt_auth_providers = {}

    # Always trust myself
    TRUSTED_ISSUERS.add(app.config['name'])

    # Install authorization check
    app.before_request(check_jwt_authorization)


def check_jwt_authorization():
    """
    Authentication handler.  Verifies jwt tokens from the session or header, if not
    already done.
    """
    current_identity = getattr(_request_ctx_stack.top, 'drift_jwt_payload', None)
    if current_identity:
        return  # authentication has already been verified

    # In case the endpoint requires no authorization, and the request does not
    # carry any authorization info as well, we will not try to verify any JWT's
    got_auth = 'Authorization' in request.headers or SESSION_COOKIE_NAME in request.cookies
    if not requires_auth(got_auth):
        return

    token, auth_type = get_auth_token_and_type()
    conf = current_app.extensions['driftconfig'].get_config()
    current_identity = verify_token(token, auth_type, conf)
    if auth_type == "JWT":
        # Cache this token for JTI identification
        cache_token(current_identity)

    # Authorization token has now been converted to a verified payload
    _request_ctx_stack.top.drift_jwt_payload = current_identity


def query_current_user():
    """
    Return the current jwt payload if the user has been authenticated
    """
    return getattr(_request_ctx_stack.top, 'drift_jwt_payload', None)


current_user = LocalProxy(query_current_user)


def requires_auth(got_auth):
    """
    returns True if the current request requires authentication
    """

    # 404 errors don't require auth
    endpoint = request.endpoint
    if not endpoint:
        return False

    if current_app.config.get("DISABLE_JWT", False):
        return False

    # check for matched endpoints
    for expr in WHITELIST_ENDPOINTS:
        if re.search(expr, request.endpoint):
            return got_auth

    # skip apis that have been decorated
    fn = current_app.view_functions.get(request.endpoint)
    if fn:
        if hasattr(fn, "view_class"):
            exempt = getattr(fn.view_class, "no_jwt_check", [])
            if request.method in exempt:
                if getattr(fn.view_class, "no_auth_header_check", False):
                    # Ignore authorization headers completely
                    return False
                else:
                    return got_auth
        else:
            # plain view function, decorated with jwt_not_required()
            if fn in _open_endpoints:
                return got_auth
    return True


def abort_unauthorized(description):
    """
    Raise an Unauthorized exception.
    """
    abort(UNAUTHORIZED, description=description)


def register_auth_provider(app, provider, handler):
    if not hasattr(app, "jwt_auth_providers"):
        app.jwt_auth_providers = {}
    app.jwt_auth_providers[provider] = handler


def requires_roles(_roles):
    """
        endpoint decorator to lock down an endpoint
        on a set of roles (comma delimitered)
    """
    def wrapper(fn):
        @wraps(fn)
        def decorator(*args, **kwargs):
            if not _roles:
                return fn(*args, **kwargs)
            current_user = query_current_user()
            if not current_user:
                abort_unauthorized("You do not have access to this resource."
                                   " It requires role '%s'" % _roles)

            required_roles = set(_roles.split(","))
            user_roles = set(current_user.get("roles", []))
            if not required_roles.intersection(user_roles):
                if not current_app.testing:
                    log.warning("User does not have the needed roles for this "
                                "call. User roles = '%s', Required roles = "
                                "'%s'. current_user = '%s'",
                                current_user.get("roles", ""),
                                _roles, repr(current_user))
                abort_unauthorized("You do not have access to this resource. "
                                   "It requires role '%s'" % _roles)
            return fn(*args, **kwargs)
        return decorator
    return wrapper


def jwt_not_required(fn):
    log.debug("Registering open endpoint in module : %s", fn.__module__)
    _open_endpoints.add(fn)
    return fn


def error_handler(error):
    desc = "{}: {}".format(error.error, error.description)
    abort(error.status_code, description=desc)


def _fix_legacy_auth(auth_info):
    """When 'auth_info' contains no 'provider' field,
    a proper 'auth_info' data will
    be constructed if either 'jwt' or 'username' and 'password' field exist.
    """
    # Auto-detect the intended authentication method and "fix" the request.
    if "jwt" in auth_info:
        auth_info['provider'] = "jwt"
        auth_info['provider_details'] = {'jwt': auth_info['jwt']}
    elif "username" in auth_info:
        username = auth_info["username"]
        if username.startswith("uuid:"):
            auth_info['provider'] = "device_id"
        else:
            auth_info['provider'] = "user+pass"
    else:
        abort_unauthorized("Bad Request. No provider specified")

    log.warning(
        "Legacy authentication detected from %s using key '%s'. Fixed for provider '%s'",
        auth_info['provider'],
        request.remote_addr,
        request.headers.get("Drift-Api-Key"),
    )

    return auth_info


def _authenticate(auth_info, conf):
    """
    Authenticate

    Does the song-and-dance against any of the supported providers and returns
    a JWT token for use in subsequent requests.
    """
    if not auth_info:
        abort_unauthorized("Bad Request. Expected json payload.")

    if "provider" not in auth_info:
        auth_info = _fix_legacy_auth(auth_info)

    provider_details = auth_info.get('provider_details')

    # TODO: Move specific auth logic outside this module.
    # Steam and Game Center. should not be in here.

    # In fact only JWT is supported by all drift based deployables. Everything else
    # is specific to drift-base.
    if auth_info['provider'] == "jwt":
        # Authenticate using a JWT. We validate the token,
        # and issue a new one based on that.
        token = provider_details['jwt']
        payload = verify_token(token, "JWT", conf)
        # Issue a JWT with same payload as the one we got
        log.debug("Authenticating using a JWT: %s", payload)
        identity = payload
    elif auth_info['provider'] == "jti":
        if provider_details and 'jti' in provider_details:
            identity = get_cached_token(provider_details['jti'])
        if not identity:
            abort_unauthorized("Bad Request. Invalid JTI.")
    else:
        identity = authenticate_with_provider(auth_info)

    if not identity or not identity.get("identity_id"):
        raise RuntimeError("authenticate must return a dict with at"
                           " least 'identity_id' field.")

    if 'service' in identity['roles']:
        expire = JWT_EXPIRATION_DELTA_FOR_SERVICES
    else:
        expire = JWT_EXPIRATION_DELTA

    ret = issue_token(identity, expire=expire)
    log.info("Authenticated: %s", identity)

    return ret


class AuthRequestSchema(ma.Schema):
    class Meta:
        strict = True
    provider = ma.fields.Str(description="Provider name")
    provider_details = ma.fields.Dict(description="Provider specific details")
    username = ma.fields.Str(description="Legacy username")
    password = ma.fields.Str(description="Legacy password")
    automatic_account_creation = ma.fields.Boolean(description="Automatically create new users", default=True)


class AuthSchema(ma.Schema):
    class Meta:
        strict = True
    token = ma.fields.String(description="Token")
    jti = ma.fields.String(description="Token id")


@bp.route('', endpoint='authentication')
class AuthApi(MethodView):
    no_auth_header_check = True
    no_jwt_check = ['GET', 'POST']

    @bp.arguments(AuthRequestSchema)
    @bp.response(AuthSchema)
    def post(self, auth_info):
        auth_info = _authenticate(auth_info, g.conf)
        return {
            'token': auth_info['token'],
            'jti': auth_info['payload']['jti'],
        }


@bp.route('/login', endpoint='login')
class AuthLoginApi(MethodView):
    no_auth_header_check = True
    no_jwt_check = ['POST']

    @bp.arguments(AuthRequestSchema)
    def post(self, auth_info):
        is_secure = request.environ.get('wsgi.url_scheme') == 'https'
        auth_info = _authenticate(auth_info, g.conf)
        host_parts = split_host(request.headers['Host'])

        # If the token is valid for any tenant the cookie must be set for cross domain.
        if auth_info['payload'].get('tenant') == '*' and host_parts['domain_name']:
            domain = '.' + host_parts['domain_name']
        else:
            domain = None

        response = make_response(redirect(url_for('root.root', _external=True)))
        response.set_cookie(
            SESSION_COOKIE_NAME,
            'JWT ' + auth_info['token'],
            expires=auth_info['payload']['exp'],
            domain=domain,
            secure=is_secure,
        )

        return response


# TODO!!! Move these endpoints elsewhere.
@bp.route('/logout', endpoint='logout')
class AuthLogoutApi(MethodView):
    no_auth_header_check = True
    no_jwt_check = ['GET', 'POST']

    def get(self):
        response = make_response(redirect(url_for('root.root', _external=True)))
        response.set_cookie(SESSION_COOKIE_NAME, '', expires=-1)
        return response

    def post(self):
        response = make_response(redirect(url_for('root.root', _external=True)))
        response.set_cookie(SESSION_COOKIE_NAME, '', expires=-1)
        return response


def authenticate_with_provider(auth_info):
    handler = current_app.jwt_auth_providers.get(auth_info['provider'])
    if not handler:
        # provide for a default handler that can deal with multiple providers
        handler = current_app.jwt_auth_providers.get("default")
    if not handler:
        abort_unauthorized(
                "Bad Request. Unknown provider '{}'. Only 'jwt' and 'jti' are "
                "supported".format(auth_info['provider'])
            )
    return handler(auth_info)


def issue_token(payload, expire=None):
    """Issue a new JWT.
    The new token has expiration of 'expire' seconds, or 24 hours by default.
    The current deployable is marked as the issuer, and the token is signed
    using the deployables private key.
    The token is cached in Redis using token key 'jti', if caching is available
    'payload' contains the custom fields to add to the tokens payload.

    The function returns a dict with the fields 'token' and 'payload' which contain
    the token and the payload embedded in the token.

    Note: This function must be called within a request context.
    """
    algorithm = JWT_ALGORITHM
    payload = dict(payload)
    payload.update(create_standard_claims(expire))

    missing_claims = list(set(JWT_REQUIRED_CLAIMS) - set(payload.keys()))
    if missing_claims:
        raise RuntimeError('Payload is missing required claims: %s' %
                           ', '.join(missing_claims))

    ts = g.conf.table_store
    public_keys = ts.get_table('public-keys')
    row = public_keys.get({
        'tier_name': g.conf.tier['tier_name'], 'deployable_name': g.conf.deployable['deployable_name']
    })
    if not row:
        raise RuntimeError("No public key found in config for tier '{}', deployable '{}'"
                           .format(g.conf.tier['tier_name'], g.conf.deployable['deployable_name']))
    key_info = row['keys'][0]  # HACK, just select the first one
    access_token = jwt.encode(payload, key_info['private_key'], algorithm=algorithm)
    cache_token(payload, expire=expire)
    log.debug("Issuing a new token: %s.", payload)
    ret = {
        'token': access_token.decode('utf-8'),
        'payload': payload,
    }
    return ret


def get_auth_token_and_type():
    auth_types_supported = ["JWT", "JTI"]
    # support Authorization in session cookie
    auth_header_value = request.cookies.get(SESSION_COOKIE_NAME)
    # override with request header if present
    auth_header_value = request.headers.get('Authorization', auth_header_value)
    if not auth_header_value:
        log.warning("No auth header for authorization required request %s",
                    request)
        abort_unauthorized("Authorization Required."
                           " Request does not contain an access token.")

    parts = auth_header_value.split()
    auth_type = parts[0].upper()

    # Legacy support
    if auth_type == "BEARER":
        auth_type = "JWT"

    if auth_type not in auth_types_supported:
        log.warning("Auth type '%s' invalid for authorization required "
                    "request %s", auth_type, request)
        abort_unauthorized(
            "Invalid authorization header. "
            "Unsupported authorization type %s. Use one of %s" %
            (auth_type, auth_types_supported)
        )
    elif len(parts) == 1:
        log.warning("Auth header contains no token for authorization"
                    " required request %s", request)
        abort_unauthorized("Invalid authorization header. Token missing")
    elif len(parts) > 2:
        log.warning("Auth header mangled for authorization"
                    " required request %s", request)
        abort_unauthorized("Invalid authorization header. "
                           "Token contains spaces")

    return parts[1], auth_type


def verify_jwt(token, conf):
    """Verify standard Json web token 'token' and return its payload."""
    algorithm = JWT_ALGORITHM
    leeway = timedelta(seconds=JWT_LEEWAY)
    verify_claims = JWT_VERIFY_CLAIMS
    required_claims = JWT_REQUIRED_CLAIMS

    options = {
        'verify_' + claim: True
        for claim in verify_claims
    }

    options.update({
        'require_' + claim: True
        for claim in required_claims
    })

    # Get issuer to see if we trust him and have the public key to verify.
    try:
        unverified_payload = jwt.decode(token,
                                        options={
                                            "verify_signature": False
                                        },
                                        algorithms=[JWT_ALGORITHM],
                                        )
    except jwt.InvalidTokenError as e:
        abort_unauthorized("Invalid token: %s" % str(e))

    issuer = unverified_payload.get("iss")
    if not issuer:
        abort_unauthorized("Invalid JWT. The 'iss' field is missing.")

    public_key = None

    if issuer in TRUSTED_ISSUERS:
        ts = conf.table_store
        public_keys = ts.get_table('public-keys')
        drift_base_key = public_keys.get(
            {'tier_name': conf.tier['tier_name'], 'deployable_name': issuer})
        if drift_base_key:
            public_key = drift_base_key['keys'][0]['public_key']

    if public_key is None:
        trusted_issuers = conf.deployable.get('jwt_trusted_issuers', [])
        for trusted_issuer in trusted_issuers:
            if trusted_issuer["iss"] == issuer:
                public_key = trusted_issuer["pub_rsa"]
                break

    if public_key is None:
        abort_unauthorized("Invalid JWT. Issuer '%s' not known or not trusted." % issuer)

    try:
        payload = jwt.decode(
            token, public_key,
            options=options,
            algorithms=[algorithm],
            leeway=leeway
        )
    except jwt.InvalidTokenError as e:
        abort_unauthorized("Invalid token: %s" % str(e))

    return payload


def verify_token(token, auth_type, conf):
    """Verifies 'token' and returns its payload."""
    if auth_type == "JTI":
        payload = get_cached_token(token)
        if not payload:
            log.info("Invalid JTI: Token '%s' not found in cache.", token)
            abort_unauthorized("Invalid JTI. Token %s does not exist." % token)
        else:
            return payload

    if auth_type != "JWT":
        abort_unauthorized("Invalid authentication type '%s'. Must be Bearer, JWT or JTI.'" % auth_type)

    payload = verify_jwt(token, conf)

    # Verify tier
    if 'tier' not in payload:
        abort_unauthorized("Invalid JWT. Token must specify 'tier'.")

    tier = payload['tier']
    tier_name = get_tier_name()
    if tier != tier_name:
        abort_unauthorized("Invalid JWT. Token is for tier '%s' but this"
            " is tier '%s'" % (tier, tier_name))

    # Verify tenant
    if 'tenant' not in payload:
        abort_unauthorized("Invalid JWT. Token must specify 'tenant'.")

    tenant = payload['tenant']
    if conf.tenant_name:
        this_tenant = conf.tenant_name['tenant_name']
    else:
        this_tenant = current_tenant_name
    if tenant != this_tenant:
        abort_unauthorized("Invalid JWT. Token is for tenant '%s' but this"
            " is tenant '%s'" % (tenant, this_tenant))

    return payload


def create_standard_claims(expire=None):
    """Return standard payload for JWT."""
    expire = expire or JWT_EXPIRATION_DELTA

    iat = datetime.utcnow()
    exp = iat + timedelta(seconds=expire)
    jti = gen_salt(20)
    iss = g.conf.deployable['deployable_name']

    standard_claims = {
        # JWT standard fields
        'iat': iat,
        'exp': exp,
        'jti': jti,
        'iss': iss,

        # Drift fields
        'tier': g.conf.tier['tier_name'],
        'tenant': g.conf.tenant_name['tenant_name'],
    }

    return standard_claims


# Cache token in Redis so that a JTI can be used instead of a JWT.  A valid JTI is implicitly trusted.
def cache_token(payload, expire=None):
    expire = expire or 86400

    # Add fudge to 'expire' so the token will live at least a little bit longer in the
    # Redis cache than the actual expiration date.
    expire += 60*10  # Ten minutes

    try:
        jti = payload['jti']
        key = "jwt:{}".format(jti)
        if hasattr(g, 'redis'):
            g.redis.set(key, json.dumps(payload), expire=expire)
            log.debug("Token cached in redis for %s seconds: %s", expire, key)
    except Exception:
        log.exception("Exception putting jwt '%s' into redis", jti)


def get_cached_token(jti):
    key = "jwt:{}".format(jti)
    data = g.redis.get(key)
    if not data:
        return None
    payload = json.loads(data)
    return payload



# https://stackoverflow.com/questions/561486/how-to-convert-an-integer-to-the-shortest-url-safe-string-in-python
ALPHABET = string.ascii_uppercase + string.ascii_lowercase + string.digits + '-_'
ALPHABET_REVERSE = dict((c, i) for (i, c) in enumerate(ALPHABET))
BASE = len(ALPHABET)
SIGN_CHARACTER = '$'


def num_encode(n):
    if n < 0:
        return SIGN_CHARACTER + num_encode(-n)
    s = []
    while True:
        n, r = divmod(n, BASE)
        s.append(ALPHABET[r])
        if n == 0: break
    return ''.join(reversed(s))


def num_decode(s):
    if s[0] == SIGN_CHARACTER:
        return -num_decode(s[1:])
    n = 0
    for c in s:
        n = n * BASE + ALPHABET_REVERSE[c]
    return n


class JwkSchema(ma.Schema):
    """JSON Web Key"""
    class Meta:
        strict = True
    kty = ma.fields.String(description="Key Type")
    use = ma.fields.String(description="Public Key Use")
    alg = ma.fields.String(description="Algorithm")
    n = ma.fields.String(description="Public Modulus")
    e = ma.fields.String(description="Public Exponent")


class JwksSchema(ma.Schema):
    class Meta:
        strict = True
    keys = ma.fields.List(ma.fields.Nested(JwkSchema))


# Expose our public key as jwks according to best practices.
# https://docs.aws.amazon.com/cognito/latest/developerguide/amazon-cognito-user-pools-using-tokens-verifying-a-jwt.html
# https://auth0.com/docs/tokens/concepts/jwks
# https://tools.ietf.org/html/rfc7517
@bpjwks.route('jwks.json')
class JWKSApi(MethodView):
    no_auth_header_check = True
    no_jwt_check = ['GET']

    @bp.response(JwksSchema)
    def get(self):
        from driftconfig.util import get_default_drift_config
        ts = get_default_drift_config()
        public_keys = ts.get_table('public-keys')
        row = public_keys.get({
            'tier_name': get_tier_name(),
            'deployable_name': current_app.config['name'],
        })
        json_web_keys = []

        if row and "keys" in row:
            for key in row["keys"]:
                pk = load_pem_public_key(key["public_key"].encode("ASCII"), backend=default_backend())

                # https://tools.ietf.org/html/rfc7517
                # MUST  "kty" (Key Type) Parameter = "RSA"
                # OPTIONAL "use" (Public Key Use) Parameter = "sig"
                # OPTIONAL "key_ops" (Key Operations) Parameter = "verify" (verify digital signature or MAC)
                # OPTIONAL "alg" (Algorithm) Parameter = "RS256"
                # OPTIONAL "kid" (Key ID) Parameter (used to hint which key was used for signing)
                jwk = {
                    "kty": "RSA",
                    "use": "sig",
                    "alg": "RS256",
                    "n": num_encode(pk.public_numbers().n),
                    "e": num_encode(pk.public_numbers().e),
                }

                json_web_keys.append(jwk)

        return {"keys": json_web_keys}


@endpoints.register
def endpoint_info(current_user):
    ret = {
        'auth': url_for("auth.authentication", _external=True),
        'auth_login': url_for("auth.login", _external=True),
        'auth_logout': url_for("auth.logout", _external=True),
        'auth_jwks': url_for("jwks.JWKSApi", _external=True),
    }

    return ret



# -*- coding: utf-8 -*-
"""
    drift - Configuration setup code
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Apply application configuration and initialize tenants.
"""
from __future__ import absolute_import
import logging
from functools import wraps

from six.moves import http_client

from werkzeug.local import LocalProxy
from flask import g, current_app
from flask import _app_ctx_stack as stack
from flask_smorest import abort

from driftconfig.relib import CHECK_INTEGRITY
from driftconfig.util import get_drift_config, get_default_drift_config, TenantNotConfigured


from drift.core.extensions.tenancy import tenant_from_hostname
from drift.utils import get_tier_name

DEFAULT_TENANT = "global"

log = logging.getLogger(__name__)


class DriftConfig(object):
    """DriftConfig Flask extension."""
    def __init__(self, app=None):
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        if not hasattr(app, 'extensions'):
            app.extensions = {}

        app.extensions['driftconfig'] = self
        app.before_request(self.before_request)
        app.after_request(self.after_request)

        # Turn off integrity checks in table stores.
        if not app.debug:
            del CHECK_INTEGRITY[:]

    def refresh(self):
        """Invalidate Redis cache, if in use, and fetch new config from source."""
        ctx = stack.top
        if ctx is not None and hasattr(ctx, 'table_store'):
            delattr(ctx, 'table_store')

    @property
    def table_store(self):
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'table_store'):
                ctx.table_store = self._get_table_store()
            return ctx.table_store

    def _get_table_store(self):
        ts = get_default_drift_config()
        return ts

    def before_request(self):
        # Add a just-in-time getter for config
        g.conf = LocalProxy(self.get_config)

    def after_request(self, response):
        # TODO: Move this logice elsewhere
        if current_app.config.get("no_response_caching", False) or not response.cache_control.max_age:
            # Turn off all caching
            response.cache_control.no_cache = True
            response.cache_control.no_store = True
            response.cache_control.max_age = 0

        return response

    def get_config(self):
        ctx = stack.top
        if ctx is not None:
            if not hasattr(ctx, 'driftconfig'):
                ctx.driftconfig = get_config_for_request()
            return ctx.driftconfig


def get_config_for_request(allow_missing_tenant=True):
    conf = get_drift_config(
        ts=current_app.extensions['driftconfig'].table_store,
        tenant_name=tenant_from_hostname,
        tier_name=get_tier_name(),
        deployable_name=current_app.config['name'],
        allow_missing_tenant=allow_missing_tenant,
    )

    if not conf.tenant_name and conf.deployable.get('jit_provision_tenants', True):
        pass
        # Make sure caller is allowed
        ##from drift.core.extensions.jwt import check_jwt_authorization, current_user
        ##check_jwt_authorization()
        ##print("current user it alst leat", current_user)

    return conf


def check_tenant_state(tenant):
    """Make sure tenant is provisioned and active."""
    if tenant['state'] != 'active':
        abort(
            http_client.BAD_REQUEST,
            description="Tenant '{}' for tier '{}' and deployable '{}' is not active, but in state '{}'.".format(
                tenant['tenant_name'], tenant['tier_name'], tenant['deployable_name'], tenant['state'])
        )

def check_tenant(f):
    """Make sure current tenant is provided, provisioned and active."""
    @wraps(f)
    def _check(*args, **kwargs):
        if not tenant_from_hostname:
            abort(
                http_client.BAD_REQUEST,
                description="No tenant specified. Please specify one using host name prefix or "
                "the environment variable DRIFT_DEFAULT_TENANT."
            )

        conf = current_app.extensions['driftconfig'].get_config()
        if not conf.tenant:
            # This will trigger a proper exception
            try:
                get_config_for_request(allow_missing_tenant=False)
                raise RuntimeError("Should not reach this.")
            except TenantNotConfigured as e:
                abort(http_client.BAD_REQUEST, description=str(e))

        check_tenant_state(g.conf.tenant)
        return f(*args, **kwargs)

    return _check



def drift_init_extension(app, **kwargs):
    DriftConfig(app)



# -*- coding: utf-8 -*-
'''
    Json Schema checker utility for Flask applications.
    ---------------------------------------------------

    Verifies json input and output against schema defs.
'''

from __future__ import absolute_import

from functools import wraps
import logging

import six
from six.moves import http_client as http_client
from six.moves import cStringIO

from flask import current_app, request, url_for
from flask import after_this_request
from flask.json import dumps, loads
from flask.wrappers import Response
from flask_smorest import abort
from click import echo

from jsonschema import validate as _validate
from jsonschema import RefResolver, FormatChecker, ValidationError

log = logging.getLogger(__name__)


class SchemaChecker(object):

    def __init__(self, app=None):
        self.app = None
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        self.app = app
        app.extensions['jsonschema'] = self

    def get_schema_for_media_type(self, media_type_name):
        for schema in self._schemas.values():
            for media_type_info in schema.get("media_types", []):
                if media_type_name == media_type_info["type_name"]:
                    schema_path = media_type_info["schema_path"]
                    resolver = RefResolver.from_schema(schema)
                    with resolver.resolving(schema_path) as entry:
                        return entry
        raise RuntimeError(
            "Schema for media type %r not found." % media_type_name)

    def get_media_type_names(self):
        """Return a flattened out list of all available media type names."""
        mtn = []
        for schema in self._schemas.values():
            for media_type_info in schema.get("media_types", []):
                mtn.append(media_type_info["type_name"])
        return mtn


def validate(instance, schema, cls=None, *args, **kwargs):
    """
    Calls jsonschema.validate() with the arguments.
    """
    format_checker = FormatChecker()
    _validate(instance, schema, cls, *args, format_checker=format_checker, **kwargs)


def get_schema_for_media_type(media_type_name):
    """
    Return schema 'media_type_name' from
    SchemaChecker instance for this Flask app.
    """
    jschema = current_app.extensions.get('jsonschema')
    if jschema is None:
        raise RuntimeError(
            'SchemaChecker instance has not been initialized for the '
            'current application: %s' % current_app)

    return jschema.get_schema_for_media_type(media_type_name)


def schema_response(media_type_name=None, schemadef=None):
    def wrapper(fn):
        @wraps(fn)
        def decorated(*args, **kwargs):
            if media_type_name:
                schema = get_schema_for_media_type(media_type_name)

                # Tag response header to contain the url.
                @after_this_request
                def add_media_type(response):
                    schema_url = url_for(
                        "schema.schema",
                        media_type_name=media_type_name,
                        _external=True
                    )
                    response.headers["X-Media-Type"] = schema_url
                    return response
            else:
                schema = schemadef

            error_level = current_app.config.get(
                'json_response_schema_validation', "log_error"
            )

            if error_level not in ["log_error", "raise_exception"]:
                return fn(*args, **kwargs)

            # Execute request and validate response
            response = fn(*args, **kwargs)

            # This looks very dodgy. Why is this logic needed?
            if isinstance(response, Response):
                if response.status_code != http_client.OK:
                    return response
                response_value = response.data
            elif isinstance(response, tuple):
                response_value = response[0]
            else:
                response_value = response

            # Push and pull the data through json serializer to get the actual
            # json object.
            json_text = dumps(response_value)
            json_object = loads(json_text)
            try:
                validate(json_object, schema)
            except ValidationError as e:
                report = generate_validation_error_report(e, json_object)
                log.error(
                    "Schema check failed for '%s'\n%s",
                    media_type_name, report
                )

                if error_level == "raise_exception":
                    raise RuntimeError(report)

            return response

        return decorated
    return wrapper


def generate_validation_error_report(e, json_object):
    """Generate a detailed report of a schema validation error."""

    # Discovering the location of the validation error is not so straight
    # forward:
    # 1. Traverse the json object using the 'path' in the validation exception
    #    and replace the offending value with a special marker.
    # 2. Pretty-print the json object indendented json text.
    # 3. Search for the special marker in the json text to find the actual
    #    line number of the error.
    # 4. Make a report by showing the error line with a context of
    #   'lines_before' and 'lines_after' number of lines on each side.

    if json_object is None:
        return "Request requires a JSON body"
    if not e.path:
        return str(e)
    marker = "3fb539de-ef7c-4e29-91f2-65c0a982f5ea"
    lines_before = 7
    lines_after = 7

    # Find the error object and replace it with the marker
    o = json_object
    for entry in list(e.path)[:-1]:
        o = o[entry]
    try:
        orig, o[e.path[0]] = o[e.path[0]], marker
    except Exception:
        # TODO: report this error
        echo("Error setting marker in schemachecker!")

    # Pretty print the object and search for the marker
    json_error = dumps(json_object, indent=4)
    io = cStringIO(json_error)

    errline = None
    for lineno, text in enumerate(io):
        if marker in text:
            errline = lineno
            break

    if errline is not None:
        # re-create report
        report = []
        json_object[e.path[0]] = orig
        json_error = dumps(json_object, indent=4)
        io = cStringIO(json_error)

        for lineno, text in enumerate(io):
            if lineno == errline:
                line_text = "{:4}: >>>".format(lineno + 1)
            else:
                line_text = "{:4}:    ".format(lineno + 1)
            report.append(line_text + text.rstrip("\n"))

        report = report[
            max(0, errline - lines_before):errline + 1 + lines_after]

        s = "Error in line {}:\n".format(errline + 1)
        s += "\n".join(report)
        s += "\n\n" + str(e)
    else:
        s = str(e)

    return s


def simple_schema_request(request_schema_properties, required=None, json_object=None):
    def wrapper(fn):
        @wraps(fn)
        def decorated(*args, **kwargs):
            ob = json_object or request.get_json() or {}
            schema = {
                "type": "object",
                "additionalProperties": False,
                "properties": request_schema_properties,
            }

            # Excuse the chattyness here. Can it be made shorter?
            if required is None:
                required_fields = list(request_schema_properties.keys())
            else:
                required_fields = required
            if required_fields:
                schema["required"] = required_fields

            check_schema(ob, schema)
            return fn(*args, **kwargs)

        return decorated
    return wrapper


def schema_request(media_type_name):
    def wrapper(fn):
        @wraps(fn)
        def decorated(*args, **kwargs):
            if isinstance(media_type_name, six.string_types):
                schema = get_schema_for_media_type(media_type_name)
            else:
                schema = media_type_name
            json_object = request.get_json()
            try:
                validate(json_object, schema)
            except ValidationError as e:
                report = generate_validation_error_report(e, json_object)
                abort(400, description=report)
            return fn(*args, **kwargs)

        return decorated
    return wrapper


def check_schema(json_object, schema, title=None):
    """Do json schema check on object and abort with 400 error if it fails."""
    try:
        validate(json_object, schema)
    except ValidationError as e:
        report = generate_validation_error_report(e, json_object)
        if title:
            report = "Schema check failed: %s\n%s" % (title, report)
        log.info("Schema validation failure\n%s", report)
        abort(400, description=report)


def drift_init_extension(app, **kwargs):
    app.jsonschema = SchemaChecker(app)



import os
from setuptools import setup, find_packages

version_file = os.path.abspath(os.path.join("drift", "VERSION"))

with open(version_file) as f:
    version = f.readlines()[0].strip()

setup(
    name='python-drift',
    version=version,
    license='MIT',
    author='Directive Games',
    author_email='info@directivegames.com',
    description='Micro-framework for SOA based applications',
    packages=find_packages(),
    url="https://github.com/dgnorth/drift",
    include_package_data=True,
    scripts=[
        'scripts/drift-admin.py',
        'scripts/sls-deploy.py',
    ],
    entry_points={'console_scripts': [
        'drift-admin = drift.management:execute_cmd',
    ]},

    install_requires=[
        'python-driftconfig',
        'Flask',
        'flask-smorest<0.23.0',
        'flask_marshmallow',
        'jsonschema',
        'pyopenssl>=17',
        'click',  # explicit requirement on the click library for echo and cmdlinge

        # Python 3 compatibility
        'six',

        # Resource module dependencies
        'SQLAlchemy',
        'Flask-SQLAlchemy',
        'marshmallow-sqlalchemy',
        'alembic',
        'psycopg2-binary>=2.7.4',
        'redis',
        'cryptography',
        'PyJWT',
        'logstash_formatter',
        'sentry-sdk[flask]',
        'blinker',
        'pytz',
    ],

    extras_require={
        'aws': [
            'boto3',
            'paramiko',
            'fabric>=2.0',
            'pyyaml',
        ],
        'test': [
            'pytest>=5.0',
            'pytest-cov',
            'codecov',
            'requests',
            'responses',
            'travispy',
        ],
    },

    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



"""
Provide consistent and up-to-date ``__version__`` strings for Python
packages.

It is easy to forget to update ``__version__`` strings when releasing
a project, and it is important that the ``__version__`` strings are
useful over the course of development even between releases,
especially if releases are infrequent.

This file provides a Version class that addresses both problems.
Version is meant to be a simple, bare-bones approach that focuses on
(a) ensuring that all declared version information matches for a
release, and (b) providing fine-grained version information via a
version control system (VCS) in between releases.  Other approaches
like versioneer.py can automate more of the process of making
releases, but they require more complex self-modifying code and code
generation techniques than the simple Python class declaration used
here.

Currently, the only VCS supported is git, but others could be added
easily.

To use Version in a project that provides a Python package named
``package`` maintained in a git repository named ``packagegit``:

1. Make the Version class available for import from your package,
   either by adding the PyPI package "param" as a dependency for your
   package, or by simply copying this file into ``package/version.py``.

2. Assuming that the current version of your package is 1.0.0, add the
   following lines to your ``package/__init__.py``::

     from param import Version
     __version__ = Version(release=(1,0,0), fpath=__file__,
                           commit="$Format:%h$", reponame="packagegit")

   (or ``from .version import Version`` if you copied the file directly.)

3. Declare the version as a string in your package's setup.py file, e.g.::

     setup_args["version"]="1.0.0"

4. In your package's setup.py script code for making a release, add a
   call to the Version.verify method. E.g.::

     setup_args = dict(name='package', version="1.0.0", ...)

     if __name__=="__main__":
          if 'upload' in sys.argv:
              import package
              package.__version__.verify(setup_args['version'])
          setup(**setup_args)

4. Tag the version of the repository to be released with a string of
   the form v*.*, i.e. ``v1.0.0`` in this example.  E.g. for git::

     git tag -a v1.0.0 -m 'Release version 1.0.0' ; git push


Now when you run ``setup.py`` to make a release via something like
``python setup.py register sdist upload``, Python will verify that the
version last tagged in the VCS is the same as what is declared in the
package and also in setup.py, aborting the release until either the
tag is corrected or the declared version is made to match the tag.
Releases installed without VCS information will then report the
declared release version.  If VCS information is available and matches
the specified repository name, then the version reported from
e.g. ``str(package.__version__)`` will provide more detailed
information about the precise VCS revision changes since the release.
See the docstring for the Version class for more detailed information.

This file is in the public domain, provided as-is, with no warranty of
any kind expressed or implied.  Anyone is free to copy, modify,
publish, use, compile, sell, or distribute it under any license, for
any purpose, commercial or non-commercial, and by any means.  The
original file is maintained at:
https://github.com/ioam/param/blob/master/param/version.py
"""


__author__ = 'Jean-Luc Stevens'

import os, subprocess
from click import echo # simple py3 porting

def run_cmd(args, cwd=None):
    echo("{}> {}".format(cwd, " ".join(args)))
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            cwd=cwd)
    output, error = (str(s.decode()).strip() for s in proc.communicate())

    if proc.returncode != 0:
        echo(error + "\n")
        raise Exception(proc.returncode, error)
    echo(output + "\n")
    return output

class Version(object):
    """
    A simple approach to Python package versioning that supports PyPI
    releases and additional information when working with version
    control. When obtaining a package from PyPI, the version returned
    is a string-formatted rendering of the supplied release tuple.
    For instance, release (1,0) tagged as ``v1.0`` in the version
    control system will return ``1.0`` for ``str(__version__)``.  Any
    number of items can be supplied in the release tuple, with either
    two or three numeric versioning levels typical.

    During development, a command like ``git describe`` will be used to
    compute the number of commits since the last version tag, the
    short commit hash, and whether the commit is dirty (has changes
    not yet committed). Version tags must start with a lowercase 'v'
    and have a period in them, e.g. v2.0, v0.9.8, v0.1a, or v0.2beta.
    Note that any non-numeric portion of the version ("a", "beta",
    etc.)  will currently be discarded for the purposes of numeric
    comparisons.

    Also note that when version control system (VCS) information is
    used, the comparison operators take into account the number of
    commits since the last version tag. This approach is often useful
    in practice to decide which version is newer for a single
    developer, but will not necessarily be reliable when comparing
    against a different fork or branch in a distributed VCS.

    For git, if you want version control information available even in
    an exported archive (e.g. a .zip file from GitHub), you can set
    the following line in the .gitattributes file of your project::

      __init__.py export-subst
    """

    def __init__(self, release=None, fpath=None, commit=None, reponame=None):
        """
        :release:  Release tuple (corresponding to the current VCS tag)
        :fpath:    Set to ``__file__`` to access version control information
        :reponame: Used to verify VCS repository name.
        """
        self.fpath = fpath
        self._expected_commit = commit
        self.expected_release = release

        self._commit = None if commit in [None, "$Format:%h$"] else commit
        self._commit_count = 0
        self._release = None
        self._dirty = False
        self.reponame = reponame

    @property
    def release(self):
        "Return the release tuple"
        return self.fetch()._release

    @property
    def commit(self):
        "A specification for this particular VCS version, e.g. a short git SHA"
        return self.fetch()._commit

    @property
    def commit_count(self):
        "Return the number of commits since the last release"
        return self.fetch()._commit_count

    @property
    def dirty(self):
        "True if there are uncommited changes, False otherwise"
        return self.fetch()._dirty


    def fetch(self):
        """
        Returns a tuple of the major version together with the
        appropriate SHA and dirty bit (for development version only).
        """
        if self._release is not None:
            return self

        self._release = self.expected_release
        if not self.fpath:
            self._commit = self._expected_commit
            return self

         # Only git right now but easily extended to SVN, Mercurial, etc.
        for cmd in ['git', 'git.cmd', 'git.exe']:
            try:
                self.git_fetch(cmd)
                break
            except EnvironmentError:
                pass
        return self


    def git_fetch(self, cmd='git'):
        try:
            if self.reponame is not None:
                # Verify this is the correct repository (since fpath could
                # be an unrelated git repository, and param could just have
                # been copied/installed into it).
                output = run_cmd([cmd, 'remote', '-v'],
                                 cwd=os.path.dirname(self.fpath))
                repo_matches = ['/' + self.reponame + '.git' ,
                                # A remote 'server:reponame.git' can also be referred
                                # to (i.e. cloned) as `server:reponame`.
                                '/' + self.reponame + ' ']
                if not any(m in output for m in repo_matches):
                    return self

            output = run_cmd(
                [cmd, 'describe', '--long', '--match', 'v*.*', '--dirty'],
                cwd=os.path.dirname(self.fpath)
            )
        except Exception as e:
            if len(e.args) > 1 and e.args[1] == 'fatal: No names found, cannot describe anything.':
                raise Exception("Cannot find any git version tags of format v*.*")
            # If there is any other error, return (release value still useful)
            else:
                echo("Git describe failed: {}".format(e))
            return self

        echo("output is {!r}".format(output))
        split = output[1:].rsplit('-', 3)
        echo("split is {}".format(split))
        #self._release = tuple(int(el) for el in split[0].split('.'))
        self._release = tuple(el for el in split[0].split('.'))
        self._commit_count = int(split[1])
        self._commit = str(split[2][1:]) # Strip out 'g' prefix ('g'=>'git')
        self._dirty = (split[-1]=='dirty')
        echo("self is {!s}".format(self))
        echo("self._release {}".format(self._release))
        echo("self._commit_count {}".format(self._commit_count))
        echo("self._commit {}".format(self._commit))
        echo("self._dirty {}".format(self._dirty))
        return self


    def __str__(self):
        """
        Version in x.y.z string format. Does not include the "v"
        prefix of the VCS version tags, for pip compatibility.

        If the commit count is non-zero or the repository is dirty,
        the string representation is equivalent to the output of::

          git describe --long --match v*.* --dirty

        (with "v" prefix removed).
        """
        if self.release is None: return 'None'
        release = '.'.join(str(el) for el in self.release)
        if self.commit_count == 0 and not self.dirty:
            return release

        dirty_status = '-dirty' if self.dirty else ''
        return '%s-%d-g%s%s' % (release, self.commit_count,
                                self.commit, dirty_status)

    def __repr__(self):
        return "Version(%r,%r,%r)" % (self.release,
                                      self.fpath if self.fpath else None,
                                      self.commit)

    def abbrev(self,dev_suffix=""):
        """
        Abbreviated string representation, optionally declaring whether it is
        a development version.
        """
        return '.'.join(str(el) for el in self.release) + \
            (dev_suffix if self.commit_count > 0 or self.dirty else "")



    def __eq__(self, other):
        """
        Two versions are considered equivalent if and only if they are
        from the same release, with the same commit count, and are not
        dirty.  Any dirty version is considered different from any
        other version, since it could potentially have any arbitrary
        changes even for the same release and commit count.
        """
        if self.dirty or other.dirty: return False
        return (self.release, self.commit_count) == (other.release, other.commit_count)

    def __gt__(self, other):
        return (self.release, self.commit_count) > (other.release, other.commit_count)

    def __lt__(self, other):
        return (self.release, self.commit_count) < (other.release, other.commit_count)


    def verify(self, string_version=None):
        """
        Check that the version information is consistent with the VCS
        before doing a release. If supplied with a string version,
        this is also checked against the current version. Should be
        called from setup.py with the declared package version before
        releasing to PyPI.
        """
        if string_version and string_version != str(self):
            raise Exception(
                "Supplied string version '%s' does not match current version '%s'." % (
                    string_version, str(self))
            )

        if self.dirty:
            raise Exception("Current working directory is dirty.")

        if self.release != self.expected_release:
            raise Exception("Declared release does not match current release tag.")

            raise Exception("Please update the VCS version tag before release.")

        if self._expected_commit not in [None, "$Format:%h$"]:
            raise Exception("Declared release does not match the VCS version tag")


# -*- coding: utf-8 -*-
import subprocess
import sys
import os

from six.moves.urllib.parse import urlparse
from click import echo


def get_branch():
    try:
        output = subprocess.check_output(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"]
        )
        output = output.decode().replace("\n", "")
        return output
    except subprocess.CalledProcessError:
        return "unknown"


def get_commit():
    try:
        output = subprocess.check_output(
            ["git", "rev-parse", "HEAD"]
        )
        output = output.decode().replace("\n", "")
        return output
    except subprocess.CalledProcessError:
        return "unknown"


def get_repo_url():
    cmd = 'git config --get remote.origin.url'
    origin_url = subprocess.check_output(cmd.split(' ')).decode()
    repository = ""
    if origin_url.startswith("http"):
        repository, _ = os.path.splitext(urlparse(origin_url).path)
    elif origin_url.startswith("git@"):
        repository = "/" + origin_url.split(":")[1].split(".")[0]
    else:
        raise Exception("Unknown origin url format")
    github_url = "https://github.com"
    return github_url + repository


def get_git_version():
    """Returns git version info bits in a dict, or None if there is no version
    to be found (i.e. no tag for instance).
    """

    # If running in Travis CI we may not have any tags as Travis does a very shallow clone
    # of the repo. We fix it (hopefully) by fetching much deeper than 50 commits back.
    if os.environ.get('TRAVIS', False):
        cmd = 'git fetch --tags --depth=2000'
        echo("Travis CI detected. Fetch deeper just in case: " + cmd)
        subprocess.call(cmd.split(' '))

    p = subprocess.Popen(
        ["git", "describe", "--tags", "--dirty", "--long"],
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    stdout, _ = p.communicate()
    stdout = str(stdout.decode())

    if p.returncode == 128:
        echo(stdout)
        return  # No tag found, probably

    if p.returncode != 0:
        echo("git command failed: {}".format(p.returncode))
        sys.exit(1)

    is_dirty = False
    stdout = stdout.strip()
    canonical = stdout
    if stdout.endswith("-dirty"):
        stdout = stdout.replace("-dirty", "")
        is_dirty = True
    tag, commits_after_tag, sha = stdout.rsplit("-", 2)
    if is_dirty or int(commits_after_tag) > 0:
        tag = canonical

    version = {
        "tag": tag,
        "commits_after_tag": int(commits_after_tag),
        "sha": sha,
        "is_dirty": is_dirty,
        "canonical": canonical,
    }

    return version


def checkout(branch_or_tag):
    """Check out a branch or tag."""
    p = subprocess.Popen(
        ["git", "checkout", branch_or_tag],
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    stdout, _ = p.communicate()
    stdout = str(stdout.decode())
    if p.returncode != 0:
        echo("git command failed: {}".format(p.returncode))
        sys.exit(1)



# -*- coding: utf-8 -*-
"""
Schames listing APIs
"""
import logging
import importlib

from six.moves import http_client

from flask import current_app, request, g
from flask.views import MethodView
from flask_smorest import Blueprint, abort
import marshmallow as ma

from driftconfig.config import TSTransaction

from driftconfig.util import get_drift_config, define_tenant, provision_tenant_resources
from driftconfig.relib import create_backend, get_store_from_url

from drift.core.extensions.schemachecker import simple_schema_request

log = logging.getLogger(__name__)
bp_provision = Blueprint("provision", "Provision", url_prefix='/provision', description="The provision API")
bp_admin = Blueprint("admin", "Admin Provision", url_prefix='/admin', description="The admin Provision API")


class AdminProvisionRequestSchema(ma.Schema):
    provisioners = ma.fields.Dict(description="The provisioners")


class AdminProvision2GetSchema(ma.Schema):
    tenant_name = ma.fields.Str(description="Name of the tenant to provision")

class AdminProvision2PostSchema(ma.Schema):
    tenant_name = ma.fields.Str(description="Name of the tenant to provision")
    preview = ma.fields.Boolean(description="Just check")


def drift_init_extension(app, api, **kwargs):
    api.register_blueprint(bp_provision)
    api.register_blueprint(bp_admin)


@bp_provision.route('', endpoint='provision_admin')
class AdminProvisionAPI(MethodView):

    no_jwt_check = ["POST"]

    @bp_provision.arguments(AdminProvisionRequestSchema)
    def post(self, args):
        """
        Provision tenant

        <ADD DESCRIPTION>
        """
        tenant_name = g.conf.tenant_name['tenant_name']
        tier_name = g.conf.tier['tier_name']

        # quick check for tenant state before downloading config
        if g.conf.tenant["state"] != "initializing":
            abort(http_client.BAD_REQUEST, message="You can only provision tenants which are in state 'initializing'. Tenant '%s' is in state '%s'" % (tenant_name, g.conf.tenant["state"]))

        args_per_provisioner = {}
        if request.json:
            for arg in args.get("provisioners", {}):
                if "provisioner" not in arg or "arguments" not in arg:
                    log.warning("Provisioner argument missing 'provisioner' or 'arguments'")
                    continue
                args_per_provisioner[arg["provisioner"]] = arg["arguments"]

        origin = g.conf.domain['origin']
        ts = get_store_from_url(origin)
        conf = get_drift_config(
            ts=ts,
            tenant_name=tenant_name,
            tier_name=tier_name,
            deployable_name=current_app.config['name']
        )

        if conf.tenant["state"] != "initializing":
            raise RuntimeError("Tenant unexpectedly found in state '%s': %s" % (conf.tenant["state"], conf.tenant))

        resources = current_app.config.get("resources")
        for module_name in resources:
            m = importlib.import_module(module_name)
            if hasattr(m, "provision"):
                provisioner_name = m.__name__.split('.')[-1]
                log.info("Provisioning '%s' for tenant '%s' on tier '%s'", provisioner_name, tenant_name, tier_name)

                args = args_per_provisioner.get(provisioner_name, {})
                m.provision(conf, args)

        # Mark the tenant as ready
        conf.tenant["state"] = "active"

        # Save out config
        log.info("Saving config to %s", origin)
        origin_backend = create_backend(origin)
        origin_backend.save_table_store(ts)

        local_origin = 'file://~/.drift/config/' + g.conf.domain['domain_name']
        log.info("Saving config to %s", local_origin)
        local_store = create_backend(local_origin)
        local_store.save_table_store(ts)

        # invalidate flask config
        current_app.extensions['driftconfig'].refresh()

        return "OK"

@bp_admin.route('/provision', endpoint='provision')
class AdminProvisionAPI2(MethodView):

    no_jwt_check = ["GET", "POST"]

    @bp_admin.arguments(AdminProvision2GetSchema)
    def get(self, args):
        """
        Get provisioned tenant

        <ADD DESCRIPTION>
        """
        tenant_name = args.get('tenant_name')
        if tenant_name:
            crit = {'tenant_name': tenant_name}
        else:
            crit = {'tier_name': g.conf.tier['tier_name']}

        tenants = g.conf.table_store.get_table('tenants').find(crit)
        return tenants

    @simple_schema_request({
        "tenant_name": {"type": "string"},
        "preview": {"type": "boolean"},
    }, required=[])
    @bp_admin.arguments(AdminProvision2PostSchema)
    def post(self, args):
        """
        Provision tenant number 2

        <ADD DESCRIPTION>
        """
        tenant_name = args.get('tenant_name') if request.json else None
        preview = args.get('preview', False) if request.json else False
        result = []

        with TSTransaction(commit_to_origin=not preview) as ts:

            if tenant_name:
                crit = {'tenant_name': tenant_name}
            else:
                crit = {'tier_name': g.conf.tier['tier_name']}

            for tenant_info in ts.get_table('tenant-names').find(crit):
                tenant_name = tenant_info['tenant_name']
                # Refresh for good measure
                define_tenant(
                    ts=ts,
                    tenant_name=tenant_name,
                    product_name=tenant_info['product_name'],
                    tier_name=tenant_info['tier_name'],
                )

                report = provision_tenant_resources(ts=ts, tenant_name=tenant_name, preview=preview)
                result.append(report)

        return result



# -*- coding: utf-8 -*-
"""
    drift - Logging setup code
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Set up logging based on config dict.

"""
from __future__ import absolute_import

import os
import logging
from logging.handlers import SysLogHandler
import logging.config
import json
import datetime
import sys
import time
import uuid
from socket import gethostname
from collections import OrderedDict
from functools import wraps
from logstash_formatter import LogstashFormatterV1

import six
from six.moves.urllib.parse import urlsplit
from flask import g, request

from drift.core.extensions.jwt import current_user
from drift.utils import get_tier_name


def get_stream_handler():
    """returns a stream handler with standard formatting for use in local development"""
    stream_handler = logging.StreamHandler()
    stream_formatter = logging.Formatter(
        fmt="%(asctime)s %(levelname)-8s %(name)-15s %(message)s"
    )
    stream_handler.setFormatter(stream_formatter)
    return stream_handler


def get_caller():
    """returns a nice string representing caller for logs
    Note: This is heavy"""
    import inspect

    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 2)
    caller = "{} ({}#{})".format(calframe[2][3], calframe[2][1], calframe[2][2])
    return caller


def get_clean_path_from_url(url):
    """extract the endpoint path from the passed in url and remove
    service information and any id's so that the endpoint path
    might be easily used in grouping.
    """
    clean_path = None
    try:
        lst = urlsplit(url)
        path = lst.path
        lst = path.split("/")
        for i, l in enumerate(lst):
            try:
                int(l)
            except ValueError:
                pass
            else:
                lst[i] = "<int>"
        # assume that the service name is the first part so we skip it
        clean_path = "/" + "/".join(lst[2:])
    except Exception:
        # Todo: should report these errors
        pass
    return clean_path


def get_log_details():
    details = OrderedDict()
    tenant_name = None
    tier_name = get_tier_name()
    remote_addr = None

    try:
        remote_addr = request.remote_addr
    except Exception:
        pass

    try:
        if hasattr(g, "conf"):
            tenant_name = (
                g.conf.tenant_name["tenant_name"] if g.conf.tenant_name else "(none)"
            )
    except RuntimeError as e:
        if "Working outside of application context" in repr(e):
            pass
        else:
            raise
    log_context = {}
    log_context["created"] = datetime.datetime.utcnow().isoformat() + "Z"
    log_context["tenant"] = tenant_name
    log_context["tier"] = tier_name
    log_context["remote_addr"] = remote_addr
    details["logger"] = log_context
    jwt_context = {}
    try:
        fields = set(
            [
                "user_id",
                "player_id",
                "roles",
                "jti",
                "user_name",
                "player_name",
                "client_id",
                "identity_id",
            ]
        )
        for k, v in current_user.items():
            if k in fields:
                key = "{}".format(k)
                jwt_context[key] = v
            if k == "roles" and v:
                jwt_context[k] = ",".join(v)
    except Exception as e:
        pass
    if jwt_context:
        details["user"] = jwt_context

    # add Drift-Log-Context" request headers to the logs
    try:
        details["client"] = json.loads(request.headers.get("Drift-Log-Context"))
    except Exception:
        pass

    return details


# Custom log record
_logRecordFactory = logging.getLogRecordFactory()


def drift_log_record_factory(*args, **kw):
    global _logRecordFactory

    logrec = _logRecordFactory(*args, **kw)
    log_details = get_log_details()
    for k, v in log_details.items():
        setattr(logrec, k, v)
    logger_fields = (
        "levelname",
        "levelno",
        "process",
        "thread",
        "name",
        "filename",
        "module",
        "funcName",
        "lineno",
    )
    for f in logger_fields:
        log_details["logger"][f] = getattr(logrec, f, None)
    try:
        correlation_id = request.correlation_id
    except Exception:
        correlation_id = None

    log_details["logger"]["correlation_id"] = correlation_id
    log_details["logger"]["created"] = datetime.datetime.utcnow().isoformat() + "Z"
    for k, v in log_details.items():
        setattr(logrec, k, v)

    return logrec


class JSONFormatter(logging.Formatter):

    """
    Format log message as JSON.
    """

    source_host = gethostname()
    log_tag = None

    def __init__(self):
        super(JSONFormatter, self).__init__()

    def formatTime(self, record, datefmt=None):
        dt = datetime.datetime.fromtimestamp(record.created)
        return dt.isoformat() + "Z"

    def get_formatted_data(self, record):

        data = OrderedDict()
        # put the timestamp first for splunk timestamp indexing
        data["timestamp"] = self.formatTime(record)
        if hasattr(record, "logger") and "tier" in record.logger:
            data["tenant"] = "{}.{}".format(
                record.logger.get("tier", None), record.logger.get("tenant", None)
            )

        field_names = "logger", "client", "user"
        data.update(
            {key: getattr(record, key) for key in field_names if hasattr(record, key)}
        )
        return data

    def format(self, record):
        data = self.get_formatted_data(record)
        json_text = json.dumps(data, default=self._json_default)
        return json_text

    def json_format(self, data):
        json_text = json.dumps(data, default=self._json_default)
        return "drift.%s: @cee: %s" % (self.log_tag, json_text)

    @staticmethod
    def _json_default(obj):
        """
        Coerce everything to strings.
        All objects representing time get output as ISO8601.
        """
        if (
            isinstance(obj, datetime.datetime)
            or isinstance(obj, datetime.date)
            or isinstance(obj, datetime.time)
        ):
            return obj.isoformat()
        else:
            return str(obj)


class ServerLogFormatter(JSONFormatter):
    log_tag = "server"

    def format(self, record):
        data = self.get_formatted_data(record)
        data["message"] = super(JSONFormatter, self).format(record)
        data["level"] = record.levelname
        try:
            data["request"] = "{} {}".format(request.method, request.url)
        except Exception:
            pass
        return self.json_format(data)


class EventLogFormatter(JSONFormatter):
    log_tag = "events"

    def format(self, record):
        data = self.get_formatted_data(record)
        data["event_name"] = super(JSONFormatter, self).format(record)
        data.update(getattr(record, "extra", {}))
        return self.json_format(data)


class ClientLogFormatter(JSONFormatter):
    log_tag = "client"

    def format(self, record):
        data = self.get_formatted_data(record)
        data.update(getattr(record, "extra", {}))
        return self.json_format(data)


def trim_logger(data):
    # remove unnecessary logger fields
    for k, v in data["logger"].copy().items():
        if k not in ["name", "tier", "tenant", "correlation_id"]:
            del data["logger"][k]


def format_request_body(key, value):
    if key == "password":
        return "*"
    else:
        # constrain the body to 64 characters per key and convert to string
        return str(value)[:64]


class RequestLogFormatter(JSONFormatter):
    log_tag = "request"

    def format(self, record):
        data = self.get_formatted_data(record)
        trim_logger(data)

        try:
            data["method"] = request.method
            data["url"] = request.url
            data["remote_addr"] = request.remote_addr
        except Exception:
            pass

        data["endpoint"] = get_clean_path_from_url(request.url)

        request_body = None
        try:
            if request.json:
                request_body = {
                    key: format_request_body(key, value)
                    for key, value in request.json.items()
                }
            else:
                request_body = request.data
        except Exception:
            pass

        if request_body:
            data["request_body"] = request_body

        try:
            data.update(getattr(record, "extra", {}))
        except Exception:
            pass

        if data.get("log_level") == 1:
            data = {
                "timestamp": data["timestamp"],
                "tenant": data["tenant"],
                "method": data["method"],
                "endpoint": data["endpoint"],
            }

        return self.json_format(data)


# Calling 'logsetup' more than once may result in multiple handlers emitting
# multiple log events for a single log call. Flagging it is a simple fix.
_setup_done = False


class StreamFormatter(logging.Formatter):
    """
    The stream formatter automatically grab the record's extra field
    and append its content to the log message
    """

    def format(self, record):
        message = super(StreamFormatter, self).format(record)
        if hasattr(record, "extra"):
            message += " | {}".format(record.extra)
        return message


def logsetup(app):
    global _setup_done
    if _setup_done:
        return
    _setup_done = True
    app.log_formatter = None

    output_format = app.config.get("LOG_FORMAT", "json").lower()
    log_level = app.config.get("LOG_LEVEL", "INFO").upper()

    if output_format == "json":
        logger = logging.getLogger()
        logger.setLevel(log_level)
        formatter = LogstashFormatterV1()
        app.log_formatter = formatter
        # make sure this is our only stream handler
        logger.handlers = []
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    else:
        logging.basicConfig(
            level=log_level, format='%(asctime)s - %(name)-14s %(levelname)-5s: %(message)s'
        )

    # if output_format == 'text':
    #     logging.basicConfig(level=log_level)
    # else:
    #     handler = logging.StreamHandler()
    #     formatter = LogstashFormatterV1()
    #     handler.setFormatter(formatter)
    #     logging.basicConfig(handlers=[handler], level=log_level)
    #     if 'logging' in app.config:
    #         logging.config.dictConfig(app.config['logging'])

    @app.before_request
    def _setup_logging():
        return setup_logging(app)


def setup_logging(app):
    """Inject a tracking identifier into the request and set up context-info
    for all debug logs
    """
    g.log_defaults = None
    request_id = request.headers.get("Request-ID", None)
    if not request_id:
        default_request_id = str(uuid.uuid4())
        request_id = request.headers.get("X-Request-ID", default_request_id)
    request.request_id = request_id

    g.log_defaults = get_log_defaults()
    if app.log_formatter:
        app.log_formatter.defaults = g.log_defaults


def get_log_defaults():
    defaults = {}
    tenant_name = None
    tier_name = get_tier_name()
    remote_addr = None

    try:
        remote_addr = request.remote_addr
    except Exception:
        pass

    try:
        if hasattr(g, 'conf'):
            tenant_name = g.conf.tenant_name['tenant_name'] if g.conf.tenant_name else '(none)'
    except RuntimeError as e:
        if "Working outside of application context" in repr(e):
            pass
        else:
            raise
    defaults["tenant"] = tenant_name
    defaults["tier"] = tier_name
    defaults["remote_addr"] = remote_addr

    jwt_context = get_user_context()

    if jwt_context:
        defaults["user"] = jwt_context

    # add Client-Log-Context" request headers to the logs
    client = None
    try:
        client = request.headers.get("Client-Log-Context", None)
        defaults["client"] = json.loads(client)
    except Exception:
        defaults["client"] = client
    defaults["request"] = {
        "request_id": request.request_id,
        "url": request.url,
        "method": request.method,
        "remote_addr": request.remote_addr,
        "path": request.path,
        "user_agent": request.headers.get('User-Agent'),
        "endpoint": get_clean_path_from_url(request.url)
    }
    defaults["request"].update(request.view_args or {})
    return defaults


def get_user_context():
    jwt_context = {}
    try:
        fields = set(["user_id", "player_id", "roles", "jti", "user_name",
                      "player_name", "client_id", "identity_id"])
        for k, v in current_user.items():
            if k in fields:
                key = "{}".format(k)
                jwt_context[key] = v
            if k == "roles" and v:
                jwt_context[k] = ",".join(v)
    except Exception:
        pass
    return jwt_context


def drift_init_extension(app, **kwargs):
    logsetup(app)


def request_log_level(level):
    def wrapper(fn):
        @wraps(fn)
        def decorated(*args, **kwargs):
            g.request_log_level = int(level)
            return fn(*args, **kwargs)

        return decorated

    return wrapper



# -*- coding: utf-8 -*-

import os
from  click import echo

from travispy import travispy
from travispy import TravisPy


def main():

    restarted = []
    building = []

    for domain in [travispy.PUBLIC, travispy.PRIVATE]:
        echo("Enumerate repos on {!r}".format(domain))
        conn = TravisPy.github_auth(os.environ['GITHUB_KEY'], domain)
        user = conn.user()
        repos = conn.repos(member=user.login)
        for repo in repos:
            if not repo.active:
                continue
            echo(u"Checking repo: {}\n{!r}".format(repo.slug, repo.description))
            try:
                build = conn.build(repo.last_build_id)
                if 'drift' in build.config.get('drift_build_trigger', []):
                    echo("Found drift project: {!r}".format(repo.slug))
                    if not build.running:
                        echo("Restarting...")
                        build.restart()
                        restarted.append(repo.slug)
                    else:
                        echo("Build is already running!")
                        building.append(repo.slug)
                else:
                    echo("Not a drift based project.")
            except Exception as e:
                echo("Can't build repo: {!r}".format(e))

            echo()

        if restarted:
            echo("Repos restarted:")
            for reponame in restarted:
                echo("\t{}".format(reponame))
        else:
            echo("No builds restarted.")

        if building:
            echo("Repos already building:")
            for reponame in building:
                echo("\t{}".format(reponame))


if __name__ == "__main__":
    main()



#
import json

from flask import Flask
from unittest import TestCase
from drift.flaskfactory import _apply_patches
from flask_smorest import Api


class DriftTestCase(TestCase):

    def __call__(self, result=None):
        """
        Does the required setup, doing it here
        means you don't have to call super.setUp
        in subclasses.
        """
        try:
            self._pre_setup()
            super(DriftTestCase, self).__call__(result)
        finally:
            self._post_teardown()

    def _pre_setup(self):
        self.app = self.create_app()
        self.client = self.app.test_client()

        self._ctx = self.app.test_request_context()
        self._ctx.push()

    def _post_teardown(self):
        if getattr(self, '_ctx', None) is not None:
            self._ctx.pop()
            del self._ctx

        if getattr(self, 'app', None) is not None:
            if getattr(self, '_orig_response_class', None) is not None:
                self.app.response_class = self._orig_response_class
            del self.app

        if hasattr(self, 'client'):
            del self.client

        if hasattr(self, 'templates'):
            del self.templates

        if hasattr(self, 'flashed_messages'):
            del self.flashed_messages

    def create_app(self):
        app = Flask(__name__)
        # apply the same kind of patching as regular factory apps get
        _apply_patches(app)
        app.config['OPENAPI_VERSION'] = "3.0.2"
        api = Api(app)
        # shitmixing this since flask-rest-api steals the 301-redirect exception
        def err(*args, **kwargs):
            pass

        api._register_error_handlers = err
        api.init_app(app)

        app.config['TESTING'] = True
        self.test_client = app.test_client()
        app.config["name"] = '{} unit test'.format(self.__class__.__name__)
        self.headers = [('Accept', 'application/json')]
        return app

    def get(self, expected_code, path, **kw):
        kw['path'] = path
        kw['headers'] = self.headers
        response = self.client.get(**kw)
        return self.assert_response(response, expected_code)

    def post(self, expected_code, path, data, **kw):
        kw['path'] = path
        kw['headers'] = self.headers
        kw['data'] = json.dumps(data)
        kw['content_type'] = 'application/json'
        response = self.client.post(**kw)
        return self.assert_response(response, expected_code)

    def delete(self, expected_code, path, **kw):
        kw['path'] = path
        kw['headers'] = self.headers
        response = self.client.delete(**kw)
        return self.assert_response(response, expected_code)

    def assert_response(self, response, expected_code):
        if response.status_code != expected_code:
            if response.headers['Content-Type'] == 'application/json':
                description = json.loads(response.data).get('description', response.data)
            else:
                description = response.data
            self.assertStatus(response, expected_code, description)
        return response

    def assertStatus(self, response, expected_code, description):
        msg = "response code:%s, expected:%s. message: %r" % (response.status_code, expected_code, description)
        self.assertEqual(response.status_code, expected_code, msg)



# -*- coding: utf-8 -*-
"""
API router target resource.

By including this resource module in a Drift app, it will be automatically configured
and included as a target for the API router.

Custom attributes for top level registration:

    api:                 <routing path prefix>
    requires_api_key:    <true|false>  Is Drift key needed or not.
"""
import logging
log = logging.getLogger(__name__)


def register_deployable(ts, deployablename, attributes):
    """
    Deployable registration callback.
    'deployablename' is from table 'deployable-names'.
    """
    pass


def register_deployable_on_tier(ts, deployable, attributes):
    """
    Deployable registration callback.
    'deployable' is from table 'deployables'.
    """
    # Add a route to this deployable.
    pk = {'tier_name': deployable['tier_name'], 'deployable_name': deployable['deployable_name']}
    row = ts.get_table('routing').get(pk)
    if row is None:
        row = ts.get_table('routing').add(pk)

    # Apply optional attributes wholesale.
    row.update(attributes)


def register_deployable_on_tenant(ts, deployable_name, tier_name, tenant_name, resource_attributes):
    pass



#!C:\Python27\python.exe
from drift import management
import argparse, os, sys
sys.dont_write_bytecode = True

if __name__ == "__main__":
    path = os.path.dirname(__file__)
    sys.path.insert(0, path)
    management.execute_cmd()
